/*
 * manager.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2011 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <sys/un.h>

#include <linux/if_addr.h>
#include <sys/stat.h>

#include "common.h"
#include "interfaces.h"
#include "ip-address.h"
#include "routes.h"
#include "bridge.h"
#include "network-inador-manager.h"
#include "dhcp_client.h"

#define COMMAND_SOCKET_PATH "/tmp/network-inador.socket"

void _manager_send_dhcp_status (ManagerClientInfo *manager_client, InterfaceDHCPClientInfo *dhcpc, gboolean is_event);

static void _manager_send_error (ManagerClientInfo *manager_client, int error, int orig_cmd) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_RESPONSE_ERROR;
	buffer[1] = orig_cmd;
	buffer[2] = error;
	
	send (manager_client->fd, buffer, 3, 0);
}

static void _manager_send_executed (ManagerClientInfo *manager_client) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_RESPONSE;
	buffer[1] = NET_INADOR_RESPONSE_EXECUTED;
	
	send (manager_client->fd, buffer, 2, 0);
}

static void _manager_send_end_command (ManagerClientInfo *manager_client, int orig_cmd) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_RESPONSE_LISTING_END;
	buffer[1] = orig_cmd;
	
	send (manager_client->fd, buffer, 2, 0);
}

static Interface * _manager_fetch_interface (ManagerClientInfo *manager_client, unsigned char *buffer, int orig_cmd) {
	Interface *iface;
	uint32_t index;
	
	memcpy (&index, buffer, 4);
	
	/* TODO: Revisar el ntohl */
	iface = _interfaces_locate_by_index (manager_client->manager->handle->interfaces, index);
	
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, orig_cmd);
		
		return NULL;
	}
	
	return iface;
}

void _manager_send_interface (ManagerClientInfo *manager_client, Interface *iface, gboolean is_event) {
	unsigned char buffer[32 + IFNAMSIZ + IFNAMSIZ];
	int name_len = strlen (iface->name);
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_IFACE_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_IFACE;
	}
	
	memcpy (&buffer[2], &iface->index, 4);
	memcpy (&buffer[6], &iface->link_type, 4);
	memcpy (&buffer[10], &iface->master_index, 4);
	memcpy (&buffer[14], &iface->mtu, 4);
	memcpy (&buffer[18], &iface->flags, 2);
	buffer[20] = iface->is_wireless;
	buffer[21] = name_len;
	memcpy (&buffer[22], iface->name, name_len);
	
	send (manager_client->fd, buffer, 22 + name_len, 0);
}

void _manager_send_interface_del (ManagerClientInfo *manager_client, uint32_t index) {
	unsigned char buffer[8];
	
	buffer[0] = NET_INADOR_TYPE_EVENT;
	buffer[1] = NET_INADOR_EVENT_IFACE_REMOVED;
	
	memcpy (&buffer[2], &index, 4);
	
	send (manager_client->fd, buffer, 6, 0);
}

void _manager_send_ip (ManagerClientInfo *manager_client, IPAddr *ip_addr, gboolean is_event) {
	unsigned char buffer[80];
	int family_size = 0;
	int label_len;
	int pos;
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_IPADDR_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_IPADDR;
	}
	
	/* Familia y máscara */
	if (ip_addr->family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (ip_addr->family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	memcpy (&buffer[2], &ip_addr->iface->index, 4);
	buffer[6] = ip_addr->family;
	buffer[7] = ip_addr->prefix;
	
	/* Los campos de bits */
	buffer[8] = 0;
	if (ip_addr->has_local) {
		buffer[8] |= 0x01;
	}
	if (ip_addr->has_brd) {
		buffer[8] |= 0x02;
	}
	
	/* Byte scope y las flags */
	buffer[9] = ip_addr->scope;
	memcpy (&buffer[10], &ip_addr->flags, 4);
	
	/* La información de marcas de tiempo de las IPs */
	memcpy (&buffer[14], &ip_addr->cacheinfo, 16);
	
	/* La dirección principal */
	memcpy (&buffer[30], &ip_addr->addr, family_size);
	pos = 30 + family_size;
	if (ip_addr->has_local) {
		memcpy (&buffer[pos], &ip_addr->local_addr, family_size);
		pos += family_size;
	}
	
	if (ip_addr->has_brd) {
		memcpy (&buffer[pos], &ip_addr->brd_addr, family_size);
		pos += family_size;
	}
	
	if (ip_addr->label[0] != 0) {
		label_len = strlen (ip_addr->label);
		buffer[pos] = label_len;
		pos++;
		
		memcpy (&buffer[pos], &ip_addr->label, label_len);
		pos += label_len;
	} else {
		buffer[pos] = 0;
		pos++;
	}
	send (manager_client->fd, buffer, pos, 0);
}

void _manager_send_ip_del (ManagerClientInfo *manager_client, IPAddr *ip_addr) {
	unsigned char buffer[80];
	int family_size = 0;
	int pos;
	
	buffer[0] = NET_INADOR_TYPE_EVENT;
	buffer[1] = NET_INADOR_EVENT_IPADDR_REMOVED;
	
	if (ip_addr->family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (ip_addr->family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	memcpy (&buffer[2], &ip_addr->iface->index, 4);
	buffer[6] = ip_addr->family;
	buffer[7] = ip_addr->prefix;
	
	buffer[8] = 0;
	if (ip_addr->has_local) {
		buffer[8] |= 0x01;
	}
	
	buffer[9] = 0;
	
	memcpy (&buffer[10], &ip_addr->addr, family_size);
	pos = 10 + family_size;
	if (ip_addr->has_local) {
		memcpy (&buffer[pos], &ip_addr->local_addr, family_size);
		pos += family_size;
	}
	
	send (manager_client->fd, buffer, pos, 0);
}

void _manager_send_list_interfaces (ManagerClientInfo *manager_client) {
	GList *g;
	Interface *iface;
	
	g = manager_client->manager->handle->interfaces;
	while (g != NULL) {
		iface = (Interface *) g->data;
		
		_manager_send_interface (manager_client, iface, FALSE);
		g = g->next;
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_IFACES);
}

void _manager_send_iface (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_GET_IFACE);
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_GET_IFACE);
	if (iface == NULL) return;
	
	_manager_send_interface (manager_client, iface, FALSE);
}

void _manager_send_list_ips (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int family;
	GList *g;
	IPAddr *ip_addr;
	Interface *iface;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_LIST_IP);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_LIST_IP);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_LIST_IP);
	if (iface == NULL) return;
	
	for (g = iface->address; g != NULL; g = g->next) {
		ip_addr = (IPAddr *) g->data;
		
		if (family != AF_UNSPEC && family != ip_addr->family) continue;
		
		_manager_send_ip (manager_client, ip_addr, FALSE);
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_IP);
}

void _manager_execute_iface_down_up (ManagerClientInfo *manager_client, int is_up, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int ret;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, (is_up ? NET_INADOR_COMMAND_IFACE_UP : NET_INADOR_COMMAND_IFACE_DOWN));
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], (is_up ? NET_INADOR_COMMAND_IFACE_UP : NET_INADOR_COMMAND_IFACE_DOWN));
	if (iface == NULL) return;
	
	if (is_up) {
		ret = interfaces_change_set_up (manager_client->manager->handle, iface->index);
	} else {
		ret = interfaces_change_set_down (manager_client->manager->handle, iface->index);
	}
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, (is_up ? NET_INADOR_COMMAND_IFACE_UP : NET_INADOR_COMMAND_IFACE_DOWN));
	}
}

void _manager_execute_iface_change_name (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int ret;
	int name_len;
	unsigned char name[IFNAMSIZ];
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
	if (iface == NULL) return;
	
	name_len = buffer[6];
	if (name_len == 0 || name_len >= IFNAMSIZ) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	if (name_len + 7 < buffer_len) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	
	memcpy (name, &buffer[7], name_len);
	if (name[0] == 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
		return;
	}
	name[name_len] = 0;
	
	ret = interfaces_change_name (manager_client->manager->handle, iface->index, name);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
	}
}

void _manager_execute_iface_change_mtu (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int ret;
	uint32_t new_mtu;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_IFACE_CHANGE_MTU);
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_IFACE_CHANGE_NAME);
	if (iface == NULL) return;
	
	memcpy (&new_mtu, &buffer[6], 4);
	//new_mtu = ntohl (new_mtu);
	
	/* TODO: Revisar el máximo de MTU que se puede poner */
	ret = interfaces_change_mtu (manager_client->manager->handle, iface->index, new_mtu);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_IFACE_CHANGE_MTU);
	}
}

void _manager_execute_clear_ips (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int family;
	Interface *iface;
	IPAddr *ip_addr;
	GList *g;
	int ret;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_CLEAR_IP);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_CLEAR_IP);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_CLEAR_IP);
	if (iface == NULL) return;
	
	ret = 0;
	for (g = iface->address; g != NULL; g = g->next) {
		ip_addr = (IPAddr *) g->data;
		
		if (family != AF_UNSPEC && family != ip_addr->family) continue;
		
		ret |= ip_address_del_ip (manager_client->manager->handle, iface->index, ip_addr);
	}
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_CLEAR_IP);
	}
}

void _manager_execute_add_ip (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	IPAddr ip_addr;
	int ret;
	int family_size, wanted_size, family;
	
	if (buffer_len < 14) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_ADD_IP);
		return;
	}
	
	family = ip_addr.family = buffer[6];
	
	if (family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_ADD_IP);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_ADD_IP);
	if (iface == NULL) return;
	
	ip_addr.prefix = buffer[7];
	if (ip_addr.family == AF_INET && (ip_addr.prefix > 32 || ip_addr.prefix < 1)) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_ADD_IP);
		return;
	} else if (ip_addr.family == AF_INET6 && (ip_addr.prefix > 128 || ip_addr.prefix < 1)) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_ADD_IP);
		return;
	}
	
	ip_addr.has_local = buffer[8] & 0x01;
	ip_addr.has_brd = (buffer[8] & 0x02) >> 1;
	
	if (ip_addr.family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (ip_addr.family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	/* Ya puedo revisar el resto de la longitud */
	wanted_size = 22 + family_size + (family_size * ip_addr.has_local) + (family_size * ip_addr.has_brd);
	if (buffer_len < wanted_size) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_ADD_IP);
		return;
	}
	
	ip_addr.scope = buffer[9];
	memcpy (&ip_addr.flags, &buffer[10], 4);
	
	/* Copiar los timestamp de tiempo válido y preferido */
	memset (&ip_addr.cacheinfo, 0, sizeof (ip_addr.cacheinfo));
	memcpy (&ip_addr.cacheinfo, &buffer[14], 8);
	
	/* Copiar las direcciones */
	memcpy (&ip_addr.addr, &buffer[22], family_size);
	wanted_size = 22 + family_size;
	
	if (ip_addr.has_local) {
		memcpy (&ip_addr.local_addr, &buffer[wanted_size], family_size);
		wanted_size += family_size;
	} else {
		memcpy (&ip_addr.local_addr, &ip_addr.addr, family_size);
	}
	
	if (ip_addr.has_brd) {
		memcpy (&ip_addr.brd_addr, &buffer[wanted_size], family_size);
		wanted_size += family_size;
	} else {
		memset (&ip_addr.brd_addr, 0, sizeof (ip_addr.brd_addr));
	}
	
	ret = ip_address_add_ip (manager_client->manager->handle, iface->index, &ip_addr);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_ADD_IP);
	}
}

void _manager_execute_delete_ip (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	IPAddr *ip_addr;
	struct_addr addr, local_addr;
	int has_local;
	int ret;
	int family_size, wanted_size, family;
	int prefix;
	
	if (buffer_len < 9) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_REMOVE_IP);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_REMOVE_IP);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_REMOVE_IP);
	if (iface == NULL) return;
	
	prefix = buffer[7];
	if (family == AF_INET && (prefix > 32 || prefix < 1)) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_REMOVE_IP);
		return;
	} else if (family == AF_INET6 && (prefix > 128 || prefix < 1)) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_REMOVE_IP);
		return;
	}
	
	has_local = buffer[8] & 0x01;
	
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	/* Ya puedo revisar el resto de la longitud */
	wanted_size = 9 + family_size + (family_size * has_local);
	if (buffer_len < wanted_size) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_REMOVE_IP);
		return;
	}
	
	memcpy (&addr, &buffer[9], family_size);
	
	if (has_local) {
		memcpy (&local_addr, &buffer[9 + family_size], family_size);
	}
	
	ip_addr = _ip_address_search_addr (iface, family, &addr, prefix, (has_local ? &local_addr : NULL));
	
	if (ip_addr == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_REMOVE_IP);
		return;
	}
	
	ret = ip_address_del_ip (manager_client->manager->handle, iface->index, ip_addr);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_REMOVE_IP);
	}
}

void _manager_execute_create_bridge (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int ret;
	int name_len;
	unsigned char name[IFNAMSIZ];
	
	if (buffer_len < 3) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_CREATE_BRIDGE);
		return;
	}
	
	name_len = buffer[2];
	if (name_len == 0 || name_len >= IFNAMSIZ) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_CREATE_BRIDGE);
		return;
	}
	
	if (name_len + 3 < buffer_len) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_CREATE_BRIDGE);
		return;
	}
	
	memcpy (name, &buffer[3], name_len);
	if (name[0] == 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_CREATE_BRIDGE);
		return;
	}
	name[name_len] = 0;
	
	ret = interfaces_bridge_create (manager_client->manager->handle, name);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_CREATE_BRIDGE);
	}
}

void _manager_execute_set_or_clear_master (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len, gboolean is_clear) {
	Interface *iface, *master;
	int ret;
	
	if (buffer_len < 6 && is_clear) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_CLEAR_MASTER);
		return;
	} else if (buffer_len < 10 && is_clear == FALSE) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_SET_MASTER);
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], (is_clear ? NET_INADOR_COMMAND_CLEAR_MASTER : NET_INADOR_COMMAND_SET_MASTER));
	
	if (is_clear == FALSE) {
		master = _manager_fetch_interface (manager_client, &buffer[6], NET_INADOR_COMMAND_SET_MASTER);
		
		ret = interfaces_bridge_set_master_interface (manager_client->manager->handle, master->index, iface->index);
	} else {
		ret = interfaces_bridge_remove_master_interface (manager_client->manager->handle, iface->index);
	}
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, (is_clear ? NET_INADOR_COMMAND_CLEAR_MASTER : NET_INADOR_COMMAND_SET_MASTER));
	}
}

static void _manager_handle_set_event_mask (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	uint32_t events;
	
	if (buffer_len < 6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_SET_EVENT_MASK);
		return;
	}
	
	memcpy (&events, &buffer[2], 4);
	printf ("___ MANAGER __ Set events de un cliente (%i), eventos: %u\n", manager_client->fd, events);
	
	manager_client->wanted_events = events;
}

#define _MANAGER_CHECK_BYTE_OR_OUT(FAMILY,BUF,POS) \
	if (BUF[POS] != 0) { \
		if (FAMILY == AF_INET && BUF[POS] != 4) { \
			_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_DHCP_CLIENT_FEED); \
			return; \
		} \
	}

static void _manager_execute_dhcp_client_feed (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int name_len, wanted;
	unsigned char name[IFNAMSIZ];
	unsigned char domain_name[256];
	Interface *iface;
	int family;
	int pos, prefix, g;
	uint32_t lease_time;
	struct_addr ip, broadcast, dhcp_server;
	struct_addr dns[8], gateways[8];
	int gw_c, dns_c;
	
	if (buffer_len < 16) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	/* Revisar la familia */
	family = buffer[2];
	
	if (family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		
		return;
	}
	
	/* Revisar el estado */
	if (buffer[3] < DHCP_CLIENT_SELECTING || buffer[3] > DHCP_CLIENT_FAILED) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	/* Revisar la longitud del nombre de la interfaz */
	name_len = buffer[4];
	if (name_len == 0 || name_len >= IFNAMSIZ) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_BAD_STRING, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	_MANAGER_CHECK_BYTE_OR_OUT(family, buffer, 5); // Dirección IP
	_MANAGER_CHECK_BYTE_OR_OUT(family, buffer, 8); // Dirección Broadcast
	_MANAGER_CHECK_BYTE_OR_OUT(family, buffer, 9); // Dirección del servidor
	
	/* Revisar la posible ruta de gateways */
	if (family == AF_INET && buffer[7] % 4 != 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	/* La lista de servidores DNS */
	if (family == AF_INET && buffer[10] % 4 != 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	wanted = 16 + name_len + buffer[5] + buffer[7] + buffer[8] + buffer[9] + buffer[10] + buffer[11];
	if (wanted < buffer_len) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	/* Recuperar el nombre de la interfaz */
	memcpy (name, &buffer[16], name_len);
	name[name_len] = 0;
	
	iface = _interfaces_locate_by_name (manager_client->manager->handle->interfaces, name);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_IFACE_INDEX, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	if (iface->dhcpc.type != IFACE_ISC_DHCLIENT && iface->dhcpc.type != IFACE_BUSYBOX_UDHCPC) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	if (iface->dhcpc.dhcp_state == DHCP_CLIENT_KILLED) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
		return;
	}
	
	/* Recuperar el resto de valores, empezando por la IP */
	if (buffer[5] != 0) {
		pos = 16 + name_len;
		memcpy (&ip, &buffer[pos], buffer[5]);
	} else {
		memset (&ip, 0, sizeof (ip));
	}
	
	/* Recuperar la ruta por defecto */
	gw_c = buffer[7] / 4;
	pos = 16 + name_len + buffer[5];
	for (g = 0; g < gw_c && g < 8; g++) {
		memcpy (&gateways[g].v4, &buffer[pos], 4);
		pos += 4;
	}
	
	/* Recuperar el broadcast */
	if (buffer[8] != 0) {
		pos = 16 + name_len + buffer[5] + buffer[7];
		memcpy (&broadcast, &buffer[pos], buffer[8]);
	} else {
		memset (&broadcast, 0, sizeof (broadcast));
	}
	
	/* Recuperar el servidor DHCP */
	if (buffer[9] != 0) {
		pos = 16 + name_len + buffer[5] + buffer[7] + buffer[8];
		memcpy (&dhcp_server, &buffer[pos], buffer[9]);
	} else {
		memset (&dhcp_server, 0, sizeof (dhcp_server));
	}
	
	/* Recuperar todos los servidores DNS */
	dns_c = buffer[10] / 4;
	pos = 16 + name_len + buffer[5] + buffer[7] + buffer[8] + buffer[9];
	for (g = 0; g < 8 && g < dns_c; g++) {
		memcpy (&dns[g].v4, &buffer[pos], 4);
		pos += 4;
	}
	
	/* Recuperar el domain name */
	domain_name[0] = 0;
	if (buffer[11] != 0) {
		pos = 16 + name_len + buffer[5] + buffer[7] + buffer[8] + buffer[9] + buffer[10];
		memcpy (domain_name, &buffer[pos], buffer[11]);
		domain_name[buffer[11]] = 0;
	}
	
	/* Copiar el lease_time */
	memcpy (&lease_time, &buffer[12], 4);
	
	/* TODO: Ejecutar aquí validaciones de estado, revisar que la IP existe, que el prefix sea válido, bla, bla, bla */
	if (family == AF_INET) {
		prefix = buffer[6];
		if (prefix > 32) {
			_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_DHCP_CLIENT_FEED);
			return;
		}
	}
	interfaces_dhcp_client_internal_feed_from_client (manager_client->manager->handle, iface->index, buffer[3], &ip, buffer[6], gateways, gw_c, &broadcast, &dhcp_server, lease_time, dns, dns_c, domain_name);
	
	/* OK */
	_manager_send_executed (manager_client);
}

static void _manager_execute_dhcp_run (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int ret;
	int family, tipo;
	uint32_t flags;
	
	if (buffer_len < 12) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RUN_DHCP);
		return;
	}
	
	family = buffer[6];
	
	//if (family != AF_INET && family != AF_INET6) {
	if (family != AF_INET) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_RUN_DHCP);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_RUN_DHCP);
	if (iface == NULL) return;
	
	tipo = buffer[7];
	if (tipo == 1) {
		tipo = IFACE_ISC_DHCLIENT;
	} else if (tipo == 2) {
		tipo = IFACE_BUSYBOX_UDHCPC;
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_RUN_DHCP);
		
		return;
	}
	
	/* TODO: Revisar las banderas */
	memcpy (&flags, &buffer[8], 4);
	
	ret = interfaces_dhcp_client_run (manager_client->manager->handle, iface->index, tipo, flags);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_RUN_DHCP);
	}
}

static void _manager_execute_dhcp_stop (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int ret;
	int family;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_STOP_DHCP);
		return;
	}
	
	family = buffer[6];
	
	//if (family != AF_INET && family != AF_INET6) {
	if (family != AF_INET) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_STOP_DHCP);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_STOP_DHCP);
	if (iface == NULL) return;
	
	ret = interfaces_dhcp_client_stop (manager_client->manager->handle, iface->index);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_STOP_DHCP);
	}
}

static void _manager_execute_dhcp_get_status (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Interface *iface;
	int family;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_GET_DHCP_STATUS);
		return;
	}
	
	family = buffer[6];
	
	//if (family != AF_INET && family != AF_INET6) {
	if (family != AF_INET) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_GET_DHCP_STATUS);
		
		return;
	}
	
	iface = _manager_fetch_interface (manager_client, &buffer[2], NET_INADOR_COMMAND_GET_DHCP_STATUS);
	if (iface == NULL) return;
	
	_manager_send_dhcp_status (manager_client, &iface->dhcpc, FALSE);
}

void _manager_send_dhcp_status (ManagerClientInfo *manager_client, InterfaceDHCPClientInfo *dhcpc, gboolean is_event) {
	Interface *iface;
	unsigned char buffer[80];
	int family_size = 0;
	int label_len;
	int pos;
	int family = AF_INET;
	struct in_addr empty;
	int g;
	
	memset (&empty, 0, sizeof (empty));
	
	iface = container_of (dhcpc, Interface, dhcpc);
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_DHCP_STATUS;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_DHCP_STATUS;
	}
	
	/* Familia */
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
	}
	
	memcpy (&buffer[2], &iface->index, 4);
	buffer[6] = family;
	buffer[7] = dhcpc->type;
	buffer[8] = dhcpc->dhcp_state;
	
	/* Prefijo */
	buffer[9] = dhcpc->prefix;
	
	/* Los campos de bits */
	buffer[10] = 0;
	if (memcmp (&dhcpc->broadcast, &empty, 4) != 0) {
		buffer[10] |= 0x01;
	}
	if (memcmp (&dhcpc->dhcp_server_ip, &empty, 4) != 0) {
		buffer[10] |= 0x02;
	}
	
	/* Cantidad de GWs */
	g = dhcpc->gw_c & 0x07;
	buffer[10] |= (g << 2);
	
	/* Cantidad de DNS */
	g = dhcpc->dns_c & 0x07;
	buffer[10] |= (g << 5);
	
	buffer[11] = strlen (dhcpc->domain_name);
	
	memcpy (&buffer[12], &dhcpc->lease_time, 4);
	
	/* La dirección principal */
	memcpy (&buffer[16], &dhcpc->ip, family_size);
	pos = 16 + family_size;
	
	for (g = 0; g < dhcpc->gw_c; g++) {
		memcpy (&buffer[pos], &dhcpc->gateways[g], family_size);
		pos += family_size;
	}
	
	if (memcmp (&dhcpc->broadcast, &empty, 4) != 0) {
		memcpy (&buffer[pos], &dhcpc->broadcast, family_size);
		pos += family_size;
	}
	
	if (memcmp (&dhcpc->dhcp_server_ip, &empty, 4) != 0) {
		memcpy (&buffer[pos], &dhcpc->dhcp_server_ip, family_size);
		pos += family_size;
	}
	
	for (g = 0; g < dhcpc->dns_c; g++) {
		memcpy (&buffer[pos], &dhcpc->dns[g], family_size);
		pos += family_size;
	}
	
	memcpy (&buffer[pos], dhcpc->domain_name, buffer[11]);
	pos += buffer[11];
	
	send (manager_client->fd, buffer, pos, 0);
}

void _manager_send_route (ManagerClientInfo *manager_client, Route *route, gboolean is_event) {
	unsigned char buffer[80];
	int family_size = 0;
	int pos, has_gw;
	struct_addr empty;
	GList *g;
	RouteNH *nexthop;
	
	memset (&empty, 0, sizeof (empty));
	
	if (is_event) {
		buffer[0] = NET_INADOR_TYPE_EVENT;
		buffer[1] = NET_INADOR_EVENT_ROUTE_ADDED;
	} else {
		buffer[0] = NET_INADOR_TYPE_RESPONSE;
		buffer[1] = NET_INADOR_RESPONSE_ROUTE;
	}
	
	/* Familia y máscara */
	if (route->family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (route->family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	buffer[2] = route->family;
	buffer[3] = route->type;
	memcpy (&buffer[4], &route->table, 4);
	buffer[8] = route->prefix;
	buffer[9] = route->protocol;
	buffer[10] = route->tos;
	buffer[11] = route->scope;
	buffer[12] = 0;
	if (memcmp (&empty, &route->prefsrc, family_size) != 0) {
		/* Tiene pref-src */
		buffer[12] |= 0x01;
	}
	
	buffer[13] = 0; /* Los next-hop se incrementarán conforme los recorra */
	memcpy (&buffer[14], &route->priority, 4);
	
	/* Copiar la Ruta de destino */
	memcpy (&buffer[18], &route->dest, family_size);
	pos = 18 + family_size;
	
	/* Copiar el pref-src, si lo tiene */
	if (buffer[12] & 0x01) {
		memcpy (&buffer[pos], &route->prefsrc, family_size);
		pos += family_size;
	}
	
	/* TODO: Recorrer todos los nexthops */
	g = route->nexthops;
	while (g != NULL) {
		nexthop = (RouteNH *) g->data;
		
		buffer[pos] = 0;
		has_gw = 0;
		if (memcmp (&nexthop->gw, &empty, family_size) != 0) {
			has_gw = 1;
			buffer[pos] |= 0x01;
		}
		pos++;
		buffer[pos++] = nexthop->nh_flags; /* Flags next-hop */
		buffer[pos++] = nexthop->nh_weight; /* Weight next-hop */
		
		buffer[pos++] = 0; /* Omitimos un byte faltante */
		
		/* La interfaz de salida */
		memcpy (&buffer[pos], &nexthop->out_index, 4);
		pos += 4;
		
		if (has_gw) {
			memcpy (&buffer[pos], &nexthop->gw, family_size);
			pos += family_size;
		}
		
		/* El contador de next-hops */
		buffer[13]++;
		
		g = g->next;
	}
	
	send (manager_client->fd, buffer, pos, 0);
}

void _manager_send_route_del (ManagerClientInfo *manager_client, Route *route) {
	unsigned char buffer[80];
	int family_size = 0;
	
	buffer[0] = NET_INADOR_TYPE_EVENT;
	buffer[1] = NET_INADOR_EVENT_ROUTE_REMOVED;
	
	/* Familia y máscara */
	if (route->family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (route->family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	buffer[2] = route->family;
	buffer[3] = route->type;
	memcpy (&buffer[4], &route->table, 4);
	buffer[8] = route->prefix;
	memcpy (&buffer[9], &route->priority, 4);
	
	/* Copiar la Ruta de destino */
	memcpy (&buffer[13], &route->dest, family_size);
	
	send (manager_client->fd, buffer, 13 + family_size, 0);
}

void _manager_send_list_routes (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int family;
	GList *g;
	Route *route;
	uint32_t table;
	
	if (buffer_len < 7) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_LIST_ROUTES);
		return;
	}
	
	family = buffer[6];
	
	if (family != AF_UNSPEC && family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_LIST_ROUTES);
		
		return;
	}
	
	memcpy (&table, &buffer[2], 4);
	
	if (family == AF_UNSPEC || family == AF_INET) {
		for (g = manager_client->manager->handle->route_v4_tables; g != NULL; g = g->next) {
			route = (Route *) g->data;
			
			if (table != 0 && route->table != table) continue;
			
			_manager_send_route (manager_client, route, FALSE);
		}
	}
	
	if (family == AF_UNSPEC || family == AF_INET6) {
		for (g = manager_client->manager->handle->route_v6_tables; g != NULL; g = g->next) {
			route = (Route *) g->data;
			
			if (table != 0 && route->table != table) continue;
			
			_manager_send_route (manager_client, route, FALSE);
		}
	}
	
	_manager_send_end_command (manager_client, NET_INADOR_COMMAND_LIST_ROUTES);
}

void _manager_execute_add_route (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	Route route;
	int ret, has_prefsrc;
	int family_size, wanted_size, family, nexthops_needed;
	int g, rem, pos;
	RouteNH *nh;
	
	if (buffer_len < 18) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_ADD_ROUTE);
		return;
	}
	
	family = buffer[2];
	
	if (family != AF_INET && family != AF_INET6) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_ADD_ROUTE);
		
		return;
	}
	
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	
	has_prefsrc = 0;
	if (buffer[12] & 0x01) {
		has_prefsrc = 1;
	}
	/* Ya puedo revisar el resto de la longitud */
	wanted_size = 18 + family_size + (family_size * has_prefsrc);
	if (buffer_len < wanted_size) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_ADD_ROUTE);
		return;
	}
	
	/* Revisar los next hops */
	g = buffer[13];
	rem = buffer_len - wanted_size;
	pos = wanted_size;
	while (g > 0) {
		if (rem < 8) {
			_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_ADD_ROUTE);
			return;
		}
		
		if (buffer[pos] & 0x01 && rem < 12) {
			/* Tiene GW */
			_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_ADD_ROUTE);
			return;
		}
		
		g--;
		if (buffer[pos] & 0x01) {
			rem -= 12;
			pos += 12;
		} else {
			rem -= 8;
			pos += 8;
		}
	}
	
	memset (&route, 0, sizeof (route));
	
	/* Ejecutar el parsing */
	route.family = family;
	route.type = buffer[3];
	memcpy (&route.table, &buffer[4], 4);
	route.prefix = buffer[8];
	route.protocol = buffer[9];
	route.tos = buffer[10];
	route.scope = buffer[11];
	memcpy (&route.priority, &buffer[14], 4);
	
	memcpy (&route.dest, &buffer[18], family_size);
	pos = 18 + family_size;
	if (has_prefsrc) {
		memcpy (&route.prefsrc, &buffer[pos], family_size);
		pos += family_size;
	}
	
	/* Recorrer cada next-hop */
	g = buffer[13];
	route.nexthops = NULL;
	while (g > 0) {
		nh = (RouteNH *) malloc (sizeof (RouteNH));
		
		memset (nh, 0, sizeof (RouteNH));
		nh->nh_flags = buffer[pos + 1];
		nh->nh_weight = buffer[pos + 2];
		memcpy (&nh->out_index, &buffer[pos + 4], 4);
		
		pos += 8;
		
		if (buffer[pos] & 0x01) {
			/* Tiene GW */
			memcpy (&nh->gw, &buffer[pos], family_size);
			pos += family_size;
		}
		route.nexthops = g_list_append (route.nexthops, nh);
		g--;
	}
	
	ret = routes_add (manager_client->manager->handle, &route);
	
	if (ret == 0) {
		/* OK */
		_manager_send_executed (manager_client);
	} else {
		_manager_send_error (manager_client, NET_INADOR_ERROR_NOT_EXECUTED, NET_INADOR_COMMAND_ADD_ROUTE);
	}
}

static gboolean _manager_client_data (GIOChannel *source, GIOCondition condition, gpointer data) {
	ManagerClientInfo *manager_client = (ManagerClientInfo *) data;
	NetworkInadorManager *manager = manager_client->manager;
	int type, command;
	Interface *iface;
	int g;
	
	unsigned char buffer[8192];
	int bytes;
	
	bytes = recv (manager_client->fd, buffer, sizeof (buffer), 0);
	
	if (bytes <= 0) {
		printf ("___ MANAGER ___ Conexión cerrada manager (%i)\n", manager_client->fd);
		/* Error de lectura o cierre */
		close (manager_client->fd);
		
		manager->connected_client_list = g_list_remove (manager->connected_client_list, manager_client);
		free (manager_client);
		
		return FALSE;
	}
	
	printf ("___ MANAGER ___ Data arrival (%i):\n", manager_client->fd);
	for (g = 0; g < bytes; g++) {
		printf ("%02hhx ", buffer[g]);
	}
	printf ("\n");
	
	/* Comando incompleto */
	if (bytes < 2) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, 0);
		
		return TRUE;
	}
	
	/* Procesar los datos del cliente */
	type = buffer[0];
	if (type != NET_INADOR_TYPE_COMMAND) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_WRONG_COMMAND, 0);
		
		return TRUE;
	}
	
	command = buffer[1];
	switch (command) {
		case NET_INADOR_COMMAND_LIST_IFACES:
			_manager_send_list_interfaces (manager_client);
			break;
		case NET_INADOR_COMMAND_GET_IFACE:
			_manager_send_iface (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_IFACE_UP:
		case NET_INADOR_COMMAND_IFACE_DOWN:
			_manager_execute_iface_down_up (manager_client, (command == NET_INADOR_COMMAND_IFACE_UP), buffer, bytes);
			break;
		case NET_INADOR_COMMAND_IFACE_CHANGE_NAME:
			_manager_execute_iface_change_name (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_IFACE_CHANGE_MTU:
			_manager_execute_iface_change_mtu (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_LIST_IP:
			_manager_send_list_ips (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_CLEAR_IP:
			_manager_execute_clear_ips (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_ADD_IP:
			_manager_execute_add_ip (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_REMOVE_IP:
			_manager_execute_delete_ip (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_SET_EVENT_MASK:
			_manager_handle_set_event_mask (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_CREATE_BRIDGE:
			_manager_execute_create_bridge (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_CLEAR_MASTER:
			_manager_execute_set_or_clear_master (manager_client, buffer, bytes, TRUE);
			break;
		case NET_INADOR_COMMAND_SET_MASTER:
			_manager_execute_set_or_clear_master (manager_client, buffer, bytes, FALSE);
			break;
		case NET_INADOR_COMMAND_DHCP_CLIENT_FEED:
			_manager_execute_dhcp_client_feed (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_RUN_DHCP:
			_manager_execute_dhcp_run (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_STOP_DHCP:
			_manager_execute_dhcp_stop (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_GET_DHCP_STATUS:
			_manager_execute_dhcp_get_status (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_LIST_ROUTES:
			_manager_send_list_routes (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_ADD_ROUTE:
			_manager_execute_add_route (manager_client, buffer, bytes);
			break;
		case NET_INADOR_COMMAND_REMOVE_ROUTE:
			
			break;
		default:
			_manager_send_error (manager_client, NET_INADOR_ERROR_WRONG_COMMAND, command);
	}
	
	return TRUE;
}

static gboolean _manager_client_connect (GIOChannel *source, GIOCondition condition, gpointer data) {
	NetworkInadorManager *manager = (NetworkInadorManager *) data;
	ManagerClientInfo *manager_client;
	int fd;
	GIOChannel *channel;
	
	fd = accept (manager->socket, NULL, NULL);
	printf ("___ MANAGER ___ Nueva conexión (%i)\n", fd);
	
	fcntl (fd, F_SETFD, 1);
	manager_client = malloc (sizeof (ManagerClientInfo));
	
	if (manager_client == NULL) {
		close (fd);
		
		return TRUE;
	}
	
	memset (manager_client, 0, sizeof (*manager_client));
	
	manager_client->fd = fd;
	
	/* Generar el vigilador */
	channel = g_io_channel_unix_new (fd);
	manager_client->source = g_io_add_watch (channel, G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP, _manager_client_data, manager_client);
	g_io_channel_unref (channel);
	manager_client->manager = manager;
	
	manager->connected_client_list = g_list_prepend (manager->connected_client_list, manager_client);
	
	return TRUE;
}

void manager_send_event_interface_add (NetworkInadorHandle *handle, Interface *iface) {
	printf ("___ MANAGER ___ Informando interfaz agregada: %s (%i)\n", iface->name, iface->index);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_interface (manager_client, iface, TRUE);
		}
	}
}

void manager_send_event_interface_del (NetworkInadorHandle *handle, uint32_t index) {
	printf ("___ MANAGER ___ Informando interfaz eliminada: %i\n", index);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_interface_del (manager_client, index);
		}
	}
}

void manager_send_event_interface_update (NetworkInadorHandle *handle, Interface *iface) {
	printf ("___ MANAGER ___ Informando interfaz actualizada: %s (%i)\n", iface->name, iface->index);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_INTERFACES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_interface (manager_client, iface, TRUE);
		}
	}
}

void manager_send_event_ip_add (NetworkInadorHandle *handle, IPAddr *ip_addr) {
	printf ("___ MANAGER ___ Informando ip agregada: %s (%i)\n", ip_addr->iface->name, ip_addr->iface->index);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_IP) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_ip (manager_client, ip_addr, TRUE);
		}
	}
}

void manager_send_event_ip_del (NetworkInadorHandle *handle, IPAddr *ip_addr) {
	printf ("___ MANAGER ___ Informando ip eliminada: %s (%i)\n", ip_addr->iface->name, ip_addr->iface->index);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_IP) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_ip_del (manager_client, ip_addr);
		}
	}
}

void manager_send_event_dhcp_change (NetworkInadorHandle *handle, InterfaceDHCPClientInfo *dhcpc) {
	Interface *iface = container_of (dhcpc, Interface, dhcpc);
	GList *g;
	ManagerClientInfo *manager_client;
	
	printf ("___ MANAGER ___ Informando cambio DHCP: %s (%i)\n", iface->name, iface->index);
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_DHCP_STATUS) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_dhcp_status (manager_client, dhcpc, TRUE);
		}
	}
}

void manager_send_event_route_add (NetworkInadorHandle *handle, Route *route) {
	char buffer_ip[1024];
	inet_ntop (route->family, &route->dest, buffer_ip, sizeof (buffer_ip));
	printf ("___ MANAGER ___ Informando ruta agregada: %s/%i [Tabla %i]\n", buffer_ip, route->prefix, route->table);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_ROUTES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_route (manager_client, route, TRUE);
		}
	}
}

void manager_send_event_route_del (NetworkInadorHandle *handle, Route *route) {
	char buffer_ip[1024];
	inet_ntop (route->family, &route->dest, buffer_ip, sizeof (buffer_ip));
	printf ("___ MANAGER ___ Informando ruta eliminada: %s/%i [Tabla %i]\n", buffer_ip, route->prefix, route->table);
	GList *g;
	ManagerClientInfo *manager_client;
	
	if (handle->manager == NULL) return;
	
	for (g = handle->manager->connected_client_list; g != NULL; g = g->next) {
		manager_client = (ManagerClientInfo *) g->data;
		
		if (manager_client->wanted_events & NET_INADOR_EVENT_MASK_ROUTES) {
			printf ("___ MANAGER ___ Informando a la conexión (%i)\n", manager_client->fd);
			_manager_send_route_del (manager_client, route);
		}
	}
}

void manager_send_event_route_update (NetworkInadorHandle *handle, Route *route) {
	manager_send_event_route_add (handle, route);
}

int manager_init (NetworkInadorHandle *handle) {
	NetworkInadorManager *manager;
	struct sockaddr_un socket_name;
	GIOChannel *channel;
	
	manager = malloc (sizeof (NetworkInadorManager));
	if (manager == NULL) {
		return -1;
	}
	
	memset (manager, 0, sizeof (*manager));
	
	manager->socket = socket (AF_UNIX, SOCK_SEQPACKET | SOCK_CLOEXEC, 0);
	
	if (manager->socket < 0) {
		perror ("Failed to create AF_UNIX socket");
		
		free (manager);
		
		return -1;
	}
	
	fcntl (manager->socket, F_SETFD, 1);
	
	memset (&socket_name, 0, sizeof (struct sockaddr_un));
	
	socket_name.sun_family = AF_UNIX;
	strncpy (socket_name.sun_path, COMMAND_SOCKET_PATH, sizeof (socket_name.sun_path) - 1);
	
	unlink (COMMAND_SOCKET_PATH);
	
	if (bind (manager->socket, (struct sockaddr *) &socket_name, sizeof (struct sockaddr_un)) < 0) {
		perror ("bind");
		
		close (manager->socket);
		manager->socket = -1;
		free (manager);
		
		return -1;
	}
	
	if (listen (manager->socket, 10) < 0) {
		perror ("listen");
		
		close (manager->socket);
		manager->socket = -1;
		free (manager);
		
		return -1;
	}
	
	/* TODO: Aplicar permisos aquí */
	chmod (COMMAND_SOCKET_PATH, 0666);
	
	channel = g_io_channel_unix_new (manager->socket);
	
	manager->source = g_io_add_watch (channel, G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP, _manager_client_connect, manager);
	g_io_channel_unref (channel);
	manager->handle = handle;
	
	handle->manager = manager;
	
	return 0;
}

